/***************************************************************************\
  "tempgen.h" - Temporary Variable Generator (Phase D)
  Revision 6; 10 Nov 2002 - 23 Jan 2003
  Copyright (C) 2002-2003  Kevin P. Albrecht

  References:
    Dragon sec 8.3
\***************************************************************************/

#ifndef TEMPGEN_H
#define TEMPGEN_H

/////////////////////////
// DEFINES //////////////
/////////////////////////

#define TEMPGEN_TEMP_PREFIX    "__"
#define TEMPGEN_TEMP_PREFIX_I  "__i"
#define TEMPGEN_TEMP_PREFIX_R  "__r"
#define TEMPGEN_TEMP_PREFIX_S  "__s"

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //
#include <stdlib.h>

// Local //
#include "global.h"
#include "error.h"
#include "table.h"

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

// External //
void tempgen_init ();
void tempgen_deinit ();
Symbol *tempgen_make (Type type);
void tempgen_check (Symbol *symbol);

// Internal //
int mCountI;       // count of temporaries; Aho's 'c' (See Aho 8.3)
Symbol **mvTempsI; // the vector holding links to the temporaries created
int mCountR;       // count of temporaries; Aho's 'c' (See Aho 8.3)
Symbol **mvTempsR; // the vector holding links to the temporaries created
int mCountS;       // count of temporaries; Aho's 'c' (See Aho 8.3)
Symbol **mvTempsS; // the vector holding links to the temporaries created


/////////////////////////
// DEFINITIONS //////////
/////////////////////////

//// tempgen_init
void tempgen_init ()
{
  int i;

  mCountI = 0;
  mCountR = 0;
  mCountS = 0;

  mvTempsI = (Symbol **) JALLOC(sizeof (Symbol *) * GLOBAL_TEMPGEN_MAX_TEMPS);
  for (i = 0; i < GLOBAL_TEMPGEN_MAX_TEMPS; i++)
  {
    mvTempsI[i] = NULL;
  }

  mvTempsR = (Symbol **) JALLOC(sizeof (Symbol *) * GLOBAL_TEMPGEN_MAX_TEMPS);
  for (i = 0; i < GLOBAL_TEMPGEN_MAX_TEMPS; i++)
  {
    mvTempsR[i] = NULL;
  }

  mvTempsS = (Symbol **) JALLOC(sizeof (Symbol *) * GLOBAL_TEMPGEN_MAX_TEMPS);
  for (i = 0; i < GLOBAL_TEMPGEN_MAX_TEMPS; i++)
  {
    mvTempsS[i] = NULL;
  }
}

//// tempgen_deinit
void tempgen_deinit ()
{
  JFREE(mvTempsI);
  JFREE(mvTempsR);
  JFREE(mvTempsS);
}

//// tempgen_make
Symbol *tempgen_make (Type type)
{
  int *pCount; // pointer to the respective mCount
  Symbol ***pTemps; // pointer to the respective mvTemps
  char *prefix;
  Symbol *symbol;
  char *tempname;

  if (type == tInt)
  {
    pCount = &mCountI;
    pTemps = &mvTempsI;
    prefix = TEMPGEN_TEMP_PREFIX_I;
  }
  else if (type == tRat)
  {
    pCount = &mCountR;
    pTemps = &mvTempsR;
    prefix = TEMPGEN_TEMP_PREFIX_R;
  }
  else if (type == tStr) 
  {
    pCount = &mCountS;
    pTemps = &mvTempsS;
    prefix = TEMPGEN_TEMP_PREFIX_S;
  }

  if ((*pCount + 1) > GLOBAL_TEMPGEN_MAX_TEMPS)
  {
    err (true, EI_MAX_TEMPS_EXCEEDED, NULL);
    erre (EFF_TERMINATE);
  }

  if ((*pTemps)[*pCount] == NULL) // need to create symbol table entry
  {
    tempname = k_new (GLOBAL_TEMPGEN_TEMP_NAME_LEN);
    k_copy (tempname, prefix);
    k_catd (tempname, conv_is(*pCount));
    symbol = scope_global_enter (tempname); //EEK// symbol = scope_global_enter (tempname);
    k_del (tempname);
    sym_set (symbol, repIdent, type, false, NULL);
    (*pTemps)[*pCount] = symbol;
    routine_addLocalVar (gCurrRout, symbol);
  }
  else // symbol already created
  {
    symbol = (*pTemps)[*pCount];
    JFREE(symbol->value);
    sym_set (symbol, repIdent, type, false, NULL);
    routine_addLocalVar (gCurrRout, symbol);
  }

  (*pCount)++;
  return symbol;
}

//// tempgen_check
//// If the symbol is a temporary, deduct one from the stack
void tempgen_check (Symbol *symbol)
{
  if (k_poss(symbol->key, TEMPGEN_TEMP_PREFIX, 0) >= 0)
  {
    // The symbol is a temporary! //
    if (symbol->type == tInt) mCountI--;
    else if (symbol->type == tRat) mCountR--;
    else if (symbol->type == tStr) mCountS--;

    if ((mCountI < 0) || (mCountR < 0) || (mCountS < 0))
    {
      err (true, EI_MIN_TEMPS_REACHED, NULL);
      erre (EFF_TERMINATE);
    }

    routine_removeLocalVar (gCurrRout, symbol);
  }
}

#endif
